"""
Helper functions for zabbixbackup.
"""
import logging
import shutil
import socket
import subprocess
from pathlib import Path
from datetime import datetime
from .tables import zabbix

logger = logging.getLogger()


def quote(s):
    """Loosely quote parameters."""
    # good enough? shlex.quote alone is adding too many extras
    # must be readable, not immediately reuseable
    r = repr(s)
    if (
        len(r) - 2 != len(s)
        or any(ch in s for ch in " ()|")
    ):
        return r

    return s


def DPopen(*args, **kwargs): # pylint: disable=C0103:invalid-name
    """Execute a command via `Popen`."""
    stderr = kwargs.get("stderr", subprocess.PIPE)
    if logger.isEnabledFor(logging.DEBUG):
        stderr = None

    kwargs["stderr"] = stderr

    return subprocess.Popen(*args, **kwargs)


def process_repr(cmd=None, env=None):
    """Loosely format a command and its environment as a string."""
    if cmd is None:
        return "None"

    rargs = map(quote, cmd)

    str_env = " ".join((
        f"{key}={quote(value)}"
        for key, value
        in env.items()))

    # Good enough
    output = ""
    if str_env:
        output += str_env + " \\\n"

    output += " ".join((
        f"\\\n    {line}" if line.startswith("-") else line for line in rargs
    ))

    return output


def run(*args, **kwargs):
    """
    Wrapper for subprocess.run.
    
    force 'text' output and returns 'stdout' as a tuple of lines
    where the last line is omitted if empty (it generally is).

    Return None on error (actual error, not the process retvalue)
    """
    if "text" not in kwargs:
        kwargs["text"] = True

    stderr = subprocess.PIPE
    if logger.isEnabledFor(logging.DEBUG):
        stderr = None

    try:
        result = subprocess.run(
            *args, **kwargs,
            check=True,
            stdout=subprocess.PIPE,
            stderr=stderr)
        out = result.stdout

    except subprocess.CalledProcessError as e:
        logger.debug("Return code %d", e.returncode)
        return None
    except FileNotFoundError:
        logger.critical("Command not found \"%s\"", args[0])
        return None

    lines = tuple(map(str.strip, out.split("\n")))
    if lines:
        if lines[-1] == "":
            return lines[:-1]

    return lines


def check_binary(*names):
    """Checks whether 'names' are all valid commands in the current shell."""

    for name in names:
        if shutil.which(name) is None:
            return False
    return True


def try_find_sockets(search: str, port):
    """Try to locate available postgresql sockets."""
    if not check_binary("netstat"):
        return tuple()

    out = run(("netstat", "-lxn"))
    sockets = []

    for line in out:
        # not perfect but it works reasonably enough
        try:
            path = Path(line.split()[-1])
        except IndexError:
            pass

        folder = str(path.parent)
        name = path.name
        if search not in folder or str(port) not in name:
            continue

        sockets.append(path)

    return tuple(sockets)


def rlookup(ipaddr):
    """Perform a reverse address lookup."""

    hosts = socket.gethostbyaddr(ipaddr) # socket.getfqdn?
    if len(hosts):
        return repr(hosts[0])
    return None


def build_compress_command(profile):
    """Helper function to prepare a compress command."""
    algo, level, extra = profile

    extension = {"xz": ".xz", "gzip": ".gz", "bzip2": ".bz2"}

    env = {}
    ext = extension[algo]

    if check_binary(algo):
        cmd = (algo, f"-{level}", ) + extra
        pipe = cmd
        return env, ext, cmd, pipe

    if check_binary("7z"):
        cmd = ("7z", "a", f"-t{algo}", )
        pipe = ("7z", "a", f"-t{algo}", "-si", )
        return env, cmd, ext, pipe

    raise NotImplementedError(f"Compression binary not available '{algo}'")


def build_tar_command(profile):
    """Helper function to prepare a tar/compress command."""
    if not check_binary("tar"):
        raise NotImplementedError("Missing tar command")

    algo, level, extra = profile

    if algo == "tar":
        return {}, ("tar", "-cf", ), ".tar"

    extension = {"xz": ".tar.xz", "gzip": ".tar.gz", "bzip2": ".tar.bz2"}
    env_map = {"xz": "XZ_OPT", "gzip": "GZIP", "bzip2": "BZIP2"}
    tar_map = {"xz": "J", "gzip": "z", "bzip2": "j"}

    compr_env_var = env_map[algo]
    compr_flags = " ".join((f"-{level}", ) + extra)
    tar_flag = tar_map[algo]

    env = {compr_env_var: compr_flags}
    ext = extension[algo]
    cmd = ("tar", f"-c{tar_flag}f", )

    return env, cmd, ext


def parse_zabbix_version(query_result):
    """Parse zabbix version from `dbversion` value."""
    raw_version = query_result[0]
    major = int(raw_version[:-6])
    minor = int(raw_version[-6:-4])
    revision = int(raw_version[-4:])

    version = f"{major}.{minor}.{revision}"

    return version, (major, minor, revision)


def create_name(args):
    """Create a suitable name for a backup."""
    dt = datetime.now().strftime("%Y%m%d-%H%M%S")
    if args.name is not None:
        return f"zabbix_{args.name}_{dt}"
    return f"zabbix_{args.host}_{dt}"


def preprocess_tables_lists(args, table_list):
    """
    Partition tables as
        ignore tables,
        schema-only tables,
        and data-tables.
    """
    logger.debug("Table list: %r", table_list)
    logger.info("Tables found: %r", len(table_list))

    tables = set(table_list)
    config = tables.intersection(zabbix.config)
    monitoring = tables.intersection(zabbix.monitoring)
    unknown = tables.difference(config, monitoring)

    logger.info("Config tables: %d", len(config))
    logger.info("Monitoring tables: %d", len(monitoring))
    logger.info("Unknown tables: %d", len(unknown))

    nodata, ignore, fail = [], [], []
    if args.monitoring == "nodata":
        nodata += monitoring

    if args.unknown == "nodata":
        nodata += unknown
    elif args.unknown == "ignore":
        ignore += unknown
    elif args.unknown == "fail":
        fail += unknown

    return sorted(ignore), sorted(nodata), sorted(fail)


def pretty_log_args(args):
    """Print arguments via 'logging.info' in a readable way"""

    keys = args._keys # pylint: disable=W0212:protected-access

    str_args = ["Arguments:"]
    for key in keys:
        value = getattr(args, key, None)
        #if value is None:
        #    continue
        if key == "passwd" and value is not None:
            value = "[omissis]"

        str_args.append(f"    {key:<24}: {value}")

    logger.info("\n".join(str_args))



"""
Helper functions for archives rotation.
"""
from pathlib import Path
from shutil import rmtree
import re
import logging
from typing import Union
from .parser_defaults import PSqlArgs, MySqlArgs

logger = logging.getLogger()


re_cfg = re.compile(r"""
    zabbix_                             # suffix
    (?P<hostorname>[^_]+?)_             # host name
    (?P<year>[0-9]{4})                  # yearmonthday-hourminutesecond
    (?P<month>[0-9]{2})                 #
    (?P<day>[0-9]{2})-                  #
    (?P<hour>[0-9]{2})                  #
    (?P<minute>[0-9]{2})                #
    (?P<second>[0-9]{2})                #
    #(?P<version>([0-9][.])+?[0-9]+?)    # zabbix version and eol
    (?P<ext>([.]tar([.](gz|xz|bz2))?))? # extension (empty if plain folder)
""", re.VERBOSE)


def rotate(args: Union[PSqlArgs, MySqlArgs]):
    """
    Perform an archive rotation keeping the last 'args.n' archives.
    """
    n = args.rotate

    if n <= 0:
        return

    name = args.name if args.name is not None else args.host

    # create a list of tuples in the form of [(datetime as int, folder)]
    # in order to being able to sort it naturally
    backups = []
    for archive in Path(".").iterdir():
        if match := re_cfg.fullmatch(archive.name):
            d = match.groupdict()
            if d["hostorname"] == name:
                int_dt = int(
                    f"{d['year']}{d['month']}{d['day']}"
                    f"{d['hour']}{d['minute']}{d['second']}"
                )
                backups.append((int_dt, archive))

    backups = sorted(backups)
    remove, keep = backups[:-n], backups[-n:]

    logger.info("Rotate backups")
    logger.info("Found %d backup/s", len(backups))
    logger.info("Deleting %d and keeping %d backup/s", len(remove), len(keep))

    for _, item in remove:
        logger.info("    deleting backup '%s'", item)
        if not args.dry_run:
            if item.is_file():
                item.unlink()
            else:
                rmtree(item, ignore_errors=True)

    for _, item in keep:
        logger.debug("    keeping backup '%s'", item)



"""
Zabbixbackup CLI entry point.

See `python3 -m zabbixbackup --help` for help.
"""

if __name__ == "__main__":
    import os
    import sys
    import logging
    from pathlib import Path
    from .parser import parse
    from .utils import create_name, pretty_log_args
    from .backup_postgre import backup_postgresql
    from .backup_mysql import backup_mysql
    from .archiver import save_files, archive
    from .rotation import rotate
    import atexit

    logger = logging.getLogger()

    # Parse and preprocess cli arguments
    args = parse(sys.argv[1:])
    scope = args.scope

    # TODO: rlookup here
    outdir = args.outdir
    abs_outdir = outdir.absolute()

    archive_dir = outdir / create_name(args)
    abs_archive_dir = archive_dir.absolute()

    # Chdir into this backup directory
    logger.debug("Create archive diretory and chdir to it: %s", archive_dir)
    prev_cwd = Path(os.getcwd()).absolute()
    atexit.register(lambda: os.chdir(prev_cwd))
    archive_dir.mkdir()
    os.chdir(archive_dir)

    log_path = Path("dump.log")

    # Create log in the destination directory
    logger.debug("Log file: %s", log_path)
    file_logger = logging.FileHandler(log_path)
    file_logger.setLevel(logging.DEBUG)
    logger.addHandler(file_logger)

    # Pretty print arguments as being parsed and processed
    pretty_log_args(args)

    if scope["dbms"] == "psql":
        status, message = backup_postgresql(args)
    elif scope["dbms"] == "mysql":
        status, message = backup_mysql(args)
    else:
        status, message = 100, "invalid dbms {scope['dbms']}"

    # exit immediately if something went wrong
    if status != 0:
        logger.fatal(message)
        sys.exit(status)

    save_files(args)

    # Detach file logger
    logger.removeHandler(file_logger)
    file_logger.close()

    # No file logging from here

    # From now on operate from backups diretory
    os.chdir(abs_outdir)
    # Archive, compress and move the backup to the final destination
    archive_path = archive(abs_archive_dir, args)

    # Rotate backups
    os.chdir(abs_outdir)
    rotate(args)

    print(archive_path)



"""
Zabbixbackup defaults configuration.
"""
from pathlib import Path
from dataclasses import dataclass, field
from typing import Optional

# Script default values
# (these are important to allow to separate user provided values during parsing)
# 'dest' name actions in parsers must be the same as in these dataclasses

# pylint: disable=W0212:protected-access
# pylint: disable=R0902:too-many-instance-attributes

@dataclass
class PSqlArgs:
    """Zabbixbackup defaults configuration for PostgreSQL."""
    read_zabbix_config: bool = False
    zabbix_config: Path      = Path("/etc/zabbix/zabbix_server.conf")

    dry_run: bool               = False
    host: str                   = "127.0.0.1"
    port: int                   = 5432
    user: str                   = "zabbix"
    passwd: str                 = None
    keeploginfile: bool         = False
    loginfile: Optional[Path]   = None
    dbname: str                 = "zabbix"
    schema: str                 = "public"
    rlookup: bool               = True
    name: Optional[str]         = None

    save_files: bool            = False
    files: Path                 = "-"

    unknown: str                = "ignore"
    monitoring: str             = "nodata"
    columns: bool               = False
    pgformat: str               = "custom"
    pgcompression: str          = None
    outdir: Path                = Path(".")
    rotate: int                 = 0

    quiet: bool                 = False
    verbose: bool               = True
    very_verbose: bool          = False
    debug: bool                 = False

    verbosity: str              = None # automatically set during parser post process
    archive: str                = "-"
    scope: dict                 = field(default_factory=dict)

PSqlArgs._keys = [
    "read_zabbix_config", "zabbix_config",
    "host", "port", "user", "passwd", "keeploginfile", "loginfile",
    "dbname", "schema", "rlookup", "name",
    "save_files", "files",
    "unknown", "monitoring",
    "columns", "pgformat", "pgcompression",
    "rotate", "outdir", "archive",
    "dry_run", "verbosity",
]


@dataclass
class MySqlArgs:
    """Zabbixbackup defaults configuration for MySQL or MariaSQL"""
    read_zabbix_config: bool    = False
    zabbix_config: Path         = Path("/etc/zabbix/zabbix_server.conf")
    read_mysql_config: bool     = False
    mysql_config: Path          = "/etc/mysql/my.cnf"

    dry_run: bool               = False
    host: str                   = "127.0.0.1"
    port: int                   = 3306
    sock: Optional[Path]        = None
    user: str                   = "zabbix"
    passwd: str                 = None
    keeploginfile: bool         = False
    loginfile: Optional[Path]   = None
    dbname: str                 = "zabbix"
    rlookup: bool               = True
    name: Optional[str]         = None

    save_files: bool            = False
    files: Path                 = "-"

    unknown: str                = "ignore"
    monitoring: str             = "nodata"
    columns: bool               = False
    mysqlcompression: str       = "gzip:6"
    outdir: Path                = Path(".")
    rotate: int                 = 0

    quiet: bool                 = False
    verbose: bool               = True
    very_verbose: bool          = False
    debug: bool                 = False

    verbosity: str              = None # automatically set during parser post process
    archive: str                = "-"
    scope: dict                 = field(default_factory=dict)

MySqlArgs._keys = [
    "read_zabbix_config", "zabbix_config", "read_mysql_config", "mysql_config",
    "host", "port", "sock", "user", "passwd", "keeploginfile", "loginfile",
    "dbname", "rlookup", "name",
    "save_files", "files",
    "unknown", "monitoring",
    "columns", "mysqlcompression",
    "rotate", "outdir", "archive",
    "dry_run", "verbosity",
]



"""
Zabbix backup utility to dump the database and save the configuration files.
"""
import logging


VERSION = 1

logging.basicConfig(format='[%(levelname)-8s] %(message)s', level=logging.DEBUG)

logger = logging.getLogger()

# Remove default logger handlers
for handler in logger.handlers:
    logger.removeHandler(handler)

# Add a logger handler for the console:
# Console logger level will be set in during argument parsing.
# An additional handler will be attached in __main__.py with
# level set to DEBUG for logfile storing
console_logger = logging.StreamHandler()
logger.addHandler(console_logger)
